/****************************************************************************
*                                                                           *
*                                                                           *
*  GTKCard, Copyright (C) 2003-2005, Identity Alliance, LLC.                *
*  Title   : P11.c                                                          *
*  Package : GTKCard                                                        *
*  Author  : Eirik A. Herskedal                                             *
*            David Corcoran                                                 *
*  Date    : 09/03/03                                                       *
*  Licence : See attached file                                              *
*                                                                           *
*                                                                           *
****************************************************************************/

#include <GTKCard.h>
#include <Mutex.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <openssl/evp.h>
#include <openssl/x509.h>
#include <openssl/pkcs12.h>
#include <openssl/err.h>

#include "pkcs11-display.h"

// Global variable for this file
pthread_mutex_t pinLock;

// Local utility functions
BOOL getFileCert(const char *file, X509 ** cert);
BOOL getFileP12(const char *file, PKCS12 ** p12);
BOOL localImportCert(X509 * cert, gchar * certID, CK_OBJECT_HANDLE * ohHandle);
CK_BYTE *keyType2Str(CK_KEY_TYPE keyType);

// List the tokens connected to the machine
BOOL listTokens(BOOL reset)
{
	// PKCS#11 variables
	CK_ULONG ulSlotCount = 0;
	CK_SLOT_ID_PTR pSlotList;
	CK_SLOT_INFO slotInfo;
	CK_TOKEN_INFO tokenInfo;
	CK_RV rv;
	char infoString[65], tokenString[33];
	CK_ULONG i;

	// GTK variables
	GList *tokenItems = NULL_PTR;

	// initialize PKCS#11
	if (globalState.p11Initialized == FALSE)
	{
		ASSERT((rv = (C_Initialize(NULL_PTR))) == CKR_OK,
			"Could not initialize PKCS#11", rv);
	}
	globalState.p11Initialized = TRUE;

	// find out if we have any readers
	// This call should fill in the slot and token arrays in the P11 lib
	ASSERT((rv = (C_GetSlotList(FALSE, NULL_PTR, &ulSlotCount))) == CKR_OK,
		"Could not get slot list", rv);

	DEBUG("There are %ld slots on this machine\n", ulSlotCount);

	ASSERT((ulSlotCount >= 1), "No slots detected", -1);

	// Now make a list over the tokens for the comboBox
	pSlotList = malloc(ulSlotCount * sizeof(CK_SLOT_ID));
	C_GetSlotList(FALSE, pSlotList, &ulSlotCount);

	for (i = 0; i < ulSlotCount; i++)
	{
		C_GetSlotInfo(pSlotList[i], &slotInfo);

		rv = C_GetTokenInfo(pSlotList[i], &tokenInfo);

		strncpy(infoString, (const char *)slotInfo.slotDescription, 64);
		infoString[64] = '\0';
		DEBUG("Slot %ld: %s\n", pSlotList[i], infoString);
		if (rv != CKR_TOKEN_NOT_PRESENT)
		{
			CK_BYTE j = 32;
			strncpy(tokenString, (char *)tokenInfo.label, 32);
			tokenString[32] = '\0';
			DEBUG("token: %s\n", tokenString);
			tokenString[32] = ' ';
			while (j > 0 && tokenString[j] == ' ')
				j--;
			if (j == 0)
			{
				sprintf(infoString, "%ld: Unrecognized token", i + 1);
			}
			else
			{
				tokenString[j + 1] = '\0';
				sprintf(infoString, "%ld: %s", pSlotList[i], tokenString);
				DEBUG("\t- %s\n", infoString);
			}
			tokenItems = g_list_append(tokenItems, strdup(infoString));
		}
		else
			DEBUG("\t- No card present\n");
	}

	if (tokenItems == NULL)
		display_main("No tokens present\n");
	else
		gtk_combo_set_popdown_strings(GTK_COMBO(globalState.tokenCombo),
			tokenItems);

	free(pSlotList);

	if (reset)
		globalState.shSession = NULL_PTR;

	return TRUE;
}

// Open a session to P#11
BOOL openSession()
{
	CK_RV rv;

	// Load the openSSL error strings and crypt algorithms
	ERR_load_crypto_strings();
	OpenSSL_add_all_algorithms();

	// initialize PKCS#11
	if (globalState.p11Initialized == FALSE)
	{
		ASSERT((rv = (C_Initialize(NULL_PTR))) == CKR_OK,
			"Could not initialize PKCS#11", rv);
	}
	globalState.p11Initialized = TRUE;

	ASSERT((rv = (C_OpenSession
				(globalState.selection, CKF_RW_SESSION | CKF_SERIAL_SESSION,
					NULL_PTR, NULL_PTR, &(globalState.shSession)))) == CKR_OK,
		"Could not open session", rv);

	DEBUG("Session opened successfully\n");

	return TRUE;
}

BOOL closeSession()
{
	CK_RV rv;

	if (globalState.shSession == NULL_PTR)
		return TRUE;

	// Unload the openSSL error strings
	ERR_free_strings();

	ASSERT((rv = (C_CloseSession(globalState.shSession))) == CKR_OK,
		"Could not close session", rv);

	globalState.shSession = NULL_PTR;

	C_Finalize(NULL);

	globalState.p11Initialized = FALSE;
	pins.adminPin.verified = FALSE;
	pins.userPin.verified = FALSE;

	DELETE();
	DISPLAY("Session closed successfully\n");

	return TRUE;
}

// Read the keys and create children in the tree
BOOL readObjects(GtkTreeIter * objectParent, GtkTreeIter * keyParent)
{
	// P11 variables
	CK_ATTRIBUTE pValueTemplate[1];
	CK_ULONG ulNumFound = 0;
	CK_RV rv;
	CK_OBJECT_HANDLE ohObjects[1];
	CK_OBJECT_CLASS ocClass;

	// Initialize the object lists
	initLists();

	// start object search
	ASSERT((rv = C_FindObjectsInit(globalState.shSession, NULL, 0)) == CKR_OK,
		"Could not initialize object search", rv);

	while (TRUE)
	{
		// do the search
		ASSERT((rv = C_FindObjects(globalState.shSession, ohObjects,
					1, &ulNumFound)) == CKR_OK,
			"Could not perform the FindObjects", rv);

		if (ulNumFound == 0)	// No more objects to read
			break;

		// Get the object class
		pValueTemplate[0].type = CKA_CLASS;
		pValueTemplate[0].pValue = &ocClass;
		pValueTemplate[0].ulValueLen = sizeof(ocClass);

		if ((rv = C_GetAttributeValue(globalState.shSession, ohObjects[0],
					pValueTemplate, 1)) != CKR_OK)
		{
			DEBUG("Error: Could not get the objects class: %s (0x%08lX)\n",
				lookup_enum(RV_T, rv), rv);
			DEBUG("Unknown object - try to interpret as a data object...\n");
			readObjectInfo(ohObjects[0]);
		}
		else
		{
			switch (ocClass)
			{
			case CKO_PRIVATE_KEY:
			case CKO_PUBLIC_KEY:
				DEBUG("Key\n");

				readKeyInfo(ohObjects[0], ocClass);

				break;
			case CKO_SECRET_KEY:
				DEBUG("Secret key\n");
				break;
			case CKO_CERTIFICATE:
				DEBUG("Cert\n");

				readCertInfo(ohObjects[0]);

				break;
			case CKO_DATA:
				DEBUG("Object\n");

				readObjectInfo(ohObjects[0]);

				break;
			case CKO_HW_FEATURE:
			case CKO_DOMAIN_PARAMETERS:
			case CKO_VENDOR_DEFINED:
				DEBUG("HW/DOMAIN/VENDOR\n");
				break;
			}
		}

	}

	C_FindObjectsFinal(globalState.shSession);

	return TRUE;
}

BOOL readObjectInfo(CK_OBJECT_HANDLE ohObject)
{
	static BOOL verified;
	char application[200];

	// GTK variables
	GtkTreeIter iter;

	// P11 variables
	CK_ATTRIBUTE pValueTemplate[2];
	CK_RV rv;
	CK_ULONG nElements, nCerts;

	// List variables
	pObjectInfo newObject;
	int nb_objects;

	newObject = malloc(sizeof(objectInfo));
	ASSERT(newObject != NULL_PTR, "Could not allocate memory", -1);

	pValueTemplate[0].type = CKA_VALUE;
	pValueTemplate[0].pValue = NULL_PTR;
	pValueTemplate[0].ulValueLen = 0;

	pValueTemplate[1].type = CKA_APPLICATION;
	pValueTemplate[1].pValue = application;
	pValueTemplate[1].ulValueLen = sizeof(application);

	rv = C_GetAttributeValue(globalState.shSession, ohObject,
		pValueTemplate, sizeof(pValueTemplate)/sizeof(pValueTemplate[0]));
	if (rv != CKR_OK)
	{
		DEBUG("Error: Could not get the objects attribute values: %s (0x%08lX)\n", lookup_enum(RV_T, rv), rv);

		//    return FALSE;

		if (verified)
			return FALSE;

		// Try to pop up a PIN dialog
		pinItemInfo.startup = TRUE;
		m_verifyPIN(globalState.app, NULL_PTR);

		// Must wait until the PIN is verified...
		DEBUG("P11 locks the mutex\n");
		lockMutex(&pinLock);
		DEBUG("P11 unlocks the mutex\n");
		unlockMutex(&pinLock);
		globalState.pause = FALSE;	// Start printing *'s again

		verified = TRUE;		// Attempted to verify the PIN
		if (pins.userPin.verified)
		{						// Make another attempt on reading object
			DEBUG("PIN was verified\n");
			if ((rv = C_GetAttributeValue(globalState.shSession,
						ohObject, pValueTemplate, 1)) != CKR_OK)
				return FALSE;
		}
		else
		{						// The pin was not verified
			DEBUG("PIN was not verified\n");
			return FALSE;
		}
	}

	/* make sure the string is null-terminated */
	application[pValueTemplate[1].ulValueLen] = '\0';

	newObject->size = pValueTemplate[0].ulValueLen;
	newObject->ohHandle = ohObject;

	append(&(lists.objects), (CK_ULONG_PTR) newObject);

	length(&(lists.objects), &nElements);
	length(&(lists.certs), &nCerts);

	gtk_tree_store_append(globalState.treeStore, &iter, &globalState.objectIter);
	gtk_tree_store_set(globalState.treeStore, &iter, TYPE_COLUMN,
		"Object", SIZE_COLUMN, newObject->size, INFORMATION_COLUMN, application,
		DATA_TYPE_COLUMN, TYPE_DATA, DATA_POINTER_COLUMN, newObject, -1);

	// one more object
	gtk_tree_model_get((GtkTreeModel *)globalState.treeStore,
		&globalState.objectIter, SIZE_COLUMN, &nb_objects, -1);
	nb_objects++;
	gtk_tree_store_set(globalState.treeStore, &globalState.objectIter,
		SIZE_COLUMN, nb_objects, -1);

	return TRUE;
}

BOOL readCertInfo(CK_OBJECT_HANDLE ohObject)
{
	// P11 variables
	CK_ATTRIBUTE pValueTemplate[2];
	CK_RV rv;
	CK_ULONG nElements, nObjects;
	CK_BYTE *tmpCert;

	// Gtk variables
	GtkTreeIter iter;

	// List variables
	pCertInfo newCert;

	// OpenSSL variables
	X509 *cert = 0;
	BIO *mem = BIO_new(BIO_s_mem());
	ASN1_INTEGER *integer;
	STACK *nlist;
	X509_NAME *name;
	int len;
	int j;
	char serialBuf[256], issuerBuf[256];
	char informationBuf[sizeof(serialBuf)+sizeof(issuerBuf)+10];
	int nb_objects;

	newCert = malloc(sizeof(certInfo));
	ASSERT(newCert != NULL_PTR, "Could not allocate memory", -1);

	pValueTemplate[0].type = CKA_ID;
	pValueTemplate[0].pValue = &(newCert->keyId);
	pValueTemplate[0].ulValueLen = sizeof(newCert->keyId);

	pValueTemplate[1].type = CKA_VALUE;
	pValueTemplate[1].pValue = NULL_PTR;
	pValueTemplate[1].ulValueLen = 0;

	ASSERT((rv = C_GetAttributeValue(globalState.shSession, ohObject,
				pValueTemplate, 2)) == CKR_OK,
		"Could not get the certificates attribute value", rv);

	newCert->size = pValueTemplate[1].ulValueLen;

	append(&(lists.certs), (CK_ULONG_PTR) newCert);

	length(&(lists.certs), &nElements);
	length(&(lists.objects), &nObjects);

	tmpCert = malloc(pValueTemplate[1].ulValueLen);

	pValueTemplate[0].type = CKA_VALUE;
	pValueTemplate[0].pValue = tmpCert;
	pValueTemplate[0].ulValueLen = pValueTemplate[1].ulValueLen;

	ASSERT((rv = C_GetAttributeValue(globalState.shSession, ohObject,
				pValueTemplate, 1)) == CKR_OK,
		"Could not get the certificates attribute values", rv);

	BIO_write(mem, tmpCert, pValueTemplate[1].ulValueLen);

	d2i_X509_bio(mem, &cert);

	if (cert == NULL)
	{
		BIO_vfree(mem);
		free(tmpCert);
		return FALSE;
	}

	name = X509_get_issuer_name(cert);
	nlist = (STACK *) name->entries;
	X509_NAME_print(mem, name, 0);
	len = BIO_read(mem, issuerBuf, 256);
	issuerBuf[len] = '\0';

	integer = X509_get_serialNumber(cert);
	for (j = 0; j < integer->length; j++)
		sprintf(&serialBuf[j * 3], "%02X%c", integer->data[j],
			(j + 1 == (int) integer->length) ? '\0' : ':');

	newCert->ohHandle = ohObject;

	g_snprintf(informationBuf, sizeof(informationBuf), "[%s : %s]", issuerBuf,
		serialBuf);
	gtk_tree_store_append(globalState.treeStore, &iter, &globalState.objectIter);
	gtk_tree_store_set(globalState.treeStore, &iter, TYPE_COLUMN,
		"Certificate", SIZE_COLUMN, newCert->size,
	   	INFORMATION_COLUMN, informationBuf, DATA_TYPE_COLUMN, TYPE_CERTIFICATE,
	   	DATA_POINTER_COLUMN, newCert, -1);

	// one more object
	gtk_tree_model_get((GtkTreeModel *)globalState.treeStore,
		&globalState.objectIter, SIZE_COLUMN, &nb_objects, -1);
	nb_objects++;
	gtk_tree_store_set(globalState.treeStore, &globalState.objectIter,
		SIZE_COLUMN, nb_objects, -1);

	BIO_vfree(mem);
	X509_free(cert);
	free(tmpCert);

	return TRUE;
}

BOOL readKeyInfo(CK_OBJECT_HANDLE ohObject, CK_OBJECT_CLASS ocClass)
{
	// P11 variables
	CK_ATTRIBUTE pValueTemplate[3];
	CK_RV rv;
	CK_BYTE *keyType;
	CK_ULONG nElements;
	CK_BYTE alignment = 0;

	// Gtk variables
	GtkTreeIter iter;

	// List variables
	pKeyInfo newKey;
	int nb_keys;

	newKey = malloc(sizeof(keyInfo));
	ASSERT(newKey != NULL_PTR, "Could not allocate memory", -1);

	pValueTemplate[0].type = CKA_ID;
	pValueTemplate[0].pValue = &(newKey->keyId);
	pValueTemplate[0].ulValueLen = sizeof(newKey->keyId);

	pValueTemplate[1].type = CKA_MODULUS;
	pValueTemplate[1].pValue = NULL;
	pValueTemplate[1].ulValueLen = 0;

	pValueTemplate[2].type = CKA_KEY_TYPE;
	pValueTemplate[2].pValue = &(newKey->keyType);
	pValueTemplate[2].ulValueLen = sizeof(newKey->keyType);

	if ((rv = C_GetAttributeValue(globalState.shSession, ohObject,
		pValueTemplate, 3)) != CKR_OK)
		DEBUG("Could not get the keys attribute value: %s (0x%08lX)\n",
			lookup_enum(RV_T, rv), rv);

	DEBUG("Len: %ld Len2: %ld Len3: %ld\n", pValueTemplate[0].ulValueLen,
		pValueTemplate[1].ulValueLen, pValueTemplate[2].ulValueLen);

	append(&(lists.keys), (CK_ULONG_PTR) newKey);

	length(&(lists.keys), &nElements);

	if (pValueTemplate[2].ulValueLen >= 0)
		keyType = keyType2Str(newKey->keyType);
	else
		keyType = (CK_BYTE *)strdup("N/A");

	if (nElements < 10)
	{
		newKey->name[0] = '0';
		alignment = 1;
	}

	newKey->class = ocClass;
	newKey->ohHandle = ohObject;

	gtk_tree_store_append(globalState.treeStore, &iter, &globalState.keyIter);
	gtk_tree_store_set(globalState.treeStore, &iter, TYPE_COLUMN,
		ocClass == CKO_PRIVATE_KEY ?  "Private key" : "Public Key",
		SIZE_COLUMN, pValueTemplate[1].ulValueLen * 8, INFORMATION_COLUMN,
		keyType, DATA_TYPE_COLUMN, TYPE_KEY, DATA_POINTER_COLUMN, newKey, -1);

	// one more key
	gtk_tree_model_get((GtkTreeModel *)globalState.treeStore,
		&globalState.keyIter, SIZE_COLUMN, &nb_keys, -1);
	nb_keys++;
	gtk_tree_store_set(globalState.treeStore, &globalState.keyIter,
		SIZE_COLUMN, nb_keys, -1);

	free(keyType);

	return TRUE;
}

BOOL login(CK_USER_TYPE userType, const char *pinCode)
{
	CK_RV rv;

	rv = C_Login(globalState.shSession, userType, (CK_CHAR_PTR)pinCode,
		strlen(pinCode));
	if (rv != CKR_OK)
	{
		if (rv == CKR_PIN_LOCKED)
		{
			DELETE();
			DISPLAY_RED("PIN code is blocked\n");
		}
		DEBUG("PIN could not be verified: %s (0x%08lX)\n",
			lookup_enum(RV_T, rv), rv);
		return FALSE;
	}

	return TRUE;
}

BOOL changePIN(const char *pinCode, const char *oldPIN)
{
	CK_RV rv;

	ASSERT((rv = C_SetPIN(globalState.shSession,
		(CK_CHAR_PTR)pinCode, strlen(pinCode),
		(CK_CHAR_PTR)oldPIN, strlen(oldPIN))) == CKR_OK,
		"PIN could not be changed\n", rv);

	return TRUE;
}

/* Create a certificate object */
BOOL importCert(gchar * fileName, const char *certID)
{
	X509 *cert;
	CK_OBJECT_HANDLE hObject = 0;
	BOOL rv;
	CK_ULONG privateKey, i = 1;
	pKeyInfo key;
	pContainer container;
	char id[21];

	DEBUG("Importing: %s\tPrivate Key: %s\n", fileName, certID);

	if (!getFileCert(fileName, &cert))
	{
		DISPLAY_RED("Could not read certificate from file\n");
		return FALSE;
	}

	privateKey = atoi(certID);
	container = lists.keys.first;
	if (container == NULL)
	{
		DISPLAY_RED("Could not locate private key\n");
		return FALSE;
	}

	while (i < privateKey && container != NULL)
	{
		container = container->next;
		i++;
	}

	if (container == NULL)
	{
		DISPLAY_RED("Could not locate private key\n");
		return FALSE;
	}

	key = container->element.key;
	if (key->class != CKO_PRIVATE_KEY)
	{
		DISPLAY_RED("Key number specified is not a private key\n");
		return FALSE;
	}

	memcpy(id, key->keyId, 20);
	id[20] = '\0';

	rv = localImportCert(cert, id, &hObject);
	readCertInfo(hObject);
	return rv;
}

/* Create new certificate and key objects from a PKCS#12 file */
BOOL importP12(gchar * fileName, const char *passwd)
{
	// P#11 variables
	CK_OBJECT_CLASS class = CKO_PRIVATE_KEY;
	CK_KEY_TYPE keyType = CKK_RSA;
	CK_BYTE_PTR subject, label;
	CK_BYTE_PTR id;
	CK_BYTE_PTR modulus;
	CK_BYTE_PTR publicExponent;
	CK_BYTE_PTR privateExponent;
	CK_BYTE_PTR prime1;
	CK_BYTE_PTR prime2;
	CK_BYTE_PTR exponent1;
	CK_BYTE_PTR exponent2;
	CK_BYTE_PTR coefficient;
	CK_BBOOL true = TRUE, false = FALSE;
	CK_OBJECT_HANDLE hObject = 0, hCert = 0;
	CK_OBJECT_HANDLE oldPubKeys[16], newPubKeys[16], newPubKey;
	CK_ATTRIBUTE pValueTemplate[3];
	CK_RV rv;
	CK_ULONG subjectSize, labelSize;
	CK_ULONG idSize;
	CK_ULONG modulusSize;
	CK_ULONG publicExponentSize;
	CK_ULONG privateExponentSize;
	CK_ULONG prime1Size;
	CK_ULONG prime2Size;
	CK_ULONG exponent1Size;
	CK_ULONG exponent2Size;
	CK_ULONG coefficientSize;

	CK_ULONG ulNumFound = 0;
	CK_OBJECT_HANDLE ohObjects[1];
	CK_OBJECT_CLASS ocClass;
	int oldPubKeyNum, newPubKeyNum, i;

	// OpenSSL variables
	PKCS12 *p12;
	EVP_PKEY *pkey = NULL;
	X509 *cert = NULL;
	RSA *key;

	// start search for public keys
	ASSERT((rv =
			C_FindObjectsInit(globalState.shSession, NULL, 0)) == CKR_OK,
		"Could not initialize object search", rv);

	oldPubKeyNum = 0;
	while (TRUE)
	{
		// do the search
		ASSERT((rv = C_FindObjects(globalState.shSession, ohObjects,
					1, &ulNumFound)) == CKR_OK,
			"Could not perform the FindObjects", rv);

		if (ulNumFound == 0)	// No more objects to read
			break;

		// Get the object class
		pValueTemplate[0].type = CKA_CLASS;
		pValueTemplate[0].pValue = &ocClass;
		pValueTemplate[0].ulValueLen = sizeof(ocClass);

		if ((rv = C_GetAttributeValue(globalState.shSession, ohObjects[0],
					pValueTemplate, 1)) != CKR_OK)
		{
			DEBUG("Error: Could not get the objects class: %s (0x%08lX)\n",
				lookup_enum(RV_T, rv), rv);
			continue;
		}

		if (ocClass == CKO_PUBLIC_KEY)
			oldPubKeys[oldPubKeyNum++] = ohObjects[0];
	}

	C_FindObjectsFinal(globalState.shSession);

	DEBUG("Importing: %s\n", fileName);

	ASSERT(getFileP12(fileName, &p12),
		"Could not get key/cert from PKCS#12 file", -1);

	rv = PKCS12_parse(p12, passwd, &pkey, &cert, NULL);

	if (!rv)
	{
		char buf[120];
		DEBUG("ERROR: Could not parse P12 structure\n");
		ERR_error_string(ERR_get_error(), buf);
		DEBUG("OpenSSL error: %s\n", buf);
		PKCS12_free(p12);
		return FALSE;
	}

	rv = localImportCert(cert, NULL, &hCert);

	if (!rv)
	{
		DEBUG("ERROR: Could not import P12 certificate\n");
		PKCS12_free(p12);
		EVP_PKEY_free(pkey);
		return FALSE;
	}

	readCertInfo(hCert);

	key = EVP_PKEY_get1_RSA(pkey);

	pValueTemplate[0].type = CKA_ID;
	pValueTemplate[0].pValue = NULL;
	pValueTemplate[0].ulValueLen = 0;

	pValueTemplate[1].type = CKA_SUBJECT;
	pValueTemplate[1].pValue = NULL;
	pValueTemplate[1].ulValueLen = 0;

	pValueTemplate[2].type = CKA_LABEL;
	pValueTemplate[2].pValue = NULL;
	pValueTemplate[2].ulValueLen = 0;

	rv = C_GetAttributeValue(globalState.shSession, hCert, pValueTemplate, 3);
	if (rv != CKR_OK)
	{
		DEBUG("Could not get the certificates attribute values lengths\n");
		PKCS12_free(p12);
		EVP_PKEY_free(pkey);
		RSA_free(key);
		return FALSE;
	}

	idSize = pValueTemplate[0].ulValueLen;
	subjectSize = pValueTemplate[1].ulValueLen;
	labelSize = pValueTemplate[2].ulValueLen;

	id = malloc(idSize);
	ASSERT(id != NULL_PTR, "Could not allocate the CKA_ID buffer\n", -1);
	subject = malloc(subjectSize);
	label = malloc(labelSize);

	pValueTemplate[0].type = CKA_ID;
	pValueTemplate[0].pValue = id;
	pValueTemplate[0].ulValueLen = idSize;

	pValueTemplate[1].type = CKA_SUBJECT;
	pValueTemplate[1].pValue = subject;
	pValueTemplate[1].ulValueLen = subjectSize;

	pValueTemplate[2].type = CKA_LABEL;
	pValueTemplate[2].pValue = label;
	pValueTemplate[2].ulValueLen = labelSize;

	rv = C_GetAttributeValue(globalState.shSession, hCert, pValueTemplate, 3);
	if (rv != CKR_OK)
	{
		DEBUG("Could not get the certificates attribute value: %s (0x%08lX)\n",
			lookup_enum(RV_T, rv), rv);
		PKCS12_free(p12);
		EVP_PKEY_free(pkey);
		RSA_free(key);
		free(id);
		free(subject);
		free(label);
		return FALSE;
	}

	modulusSize = BN_num_bytes(key->n);
	modulus = malloc(modulusSize);
	BN_bn2bin(key->n, modulus);

	publicExponentSize = BN_num_bytes(key->e);
	publicExponent = malloc(publicExponentSize);
	BN_bn2bin(key->e, publicExponent);

	privateExponentSize = BN_num_bytes(key->d);
	privateExponent = malloc(privateExponentSize);
	BN_bn2bin(key->d, privateExponent);

	prime1Size = BN_num_bytes(key->p);
	prime1 = malloc(prime1Size);
	BN_bn2bin(key->p, prime1);

	prime2Size = BN_num_bytes(key->q);
	prime2 = malloc(prime2Size);
	BN_bn2bin(key->q, prime2);

	exponent1Size = BN_num_bytes(key->dmp1);
	exponent1 = malloc(exponent1Size);
	BN_bn2bin(key->dmp1, exponent1);

	exponent2Size = BN_num_bytes(key->dmq1);
	exponent2 = malloc(exponent2Size);
	BN_bn2bin(key->dmq1, exponent2);

	coefficientSize = BN_num_bytes(key->iqmp);
	coefficient = malloc(coefficientSize);
	BN_bn2bin(key->iqmp, coefficient);

	CK_ATTRIBUTE template[] = {
		{CKA_CLASS, &class, sizeof(class)}
		,
		{CKA_KEY_TYPE, &keyType, sizeof(keyType)}
		,
		{CKA_TOKEN, &true, sizeof(true)}
		,
		{CKA_SENSITIVE, &true, sizeof(true)}
		,
		{CKA_DECRYPT, &true, sizeof(true)}
		,
		{CKA_SIGN, &true, sizeof(true)}
		,
		{CKA_EXTRACTABLE, &false, sizeof(false)}
		,
		{CKA_SIGN_RECOVER, &false, sizeof(false)}
		,
		{CKA_DERIVE, &false, sizeof(false)}
		,
		{CKA_MODIFIABLE, &false, sizeof(false)}
		,
		{CKA_UNWRAP, &true, sizeof(true)}
		,
		{CKA_PRIVATE, &true, sizeof(true)}
		,
		{CKA_NEVER_EXTRACTABLE, &true, sizeof(true)}
		,
		{CKA_ALWAYS_SENSITIVE, &true, sizeof(true)}
		,
		{CKA_MODULUS, modulus, modulusSize}
		,
		{CKA_PUBLIC_EXPONENT, publicExponent, publicExponentSize}
		,
		{CKA_PRIVATE_EXPONENT, privateExponent, privateExponentSize}
		,
		{CKA_PRIME_1, prime1, prime1Size}
		,
		{CKA_PRIME_2, prime2, prime2Size}
		,
		{CKA_EXPONENT_1, exponent1, exponent1Size}
		,
		{CKA_EXPONENT_2, exponent2, exponent2Size}
		,
		{CKA_COEFFICIENT, coefficient, coefficientSize}
	};

	if ((rv = C_CreateObject(globalState.shSession, template,
				sizeof(template) / sizeof(CK_ATTRIBUTE), &hObject)) != CKR_OK)
		DEBUG("Could not create private key, error: %s (0x%08lX)\n",
			lookup_enum(RV_T, rv), rv);
	else
	{
		{
			CK_ATTRIBUTE pValueTemplate[] = {
				{CKA_ID, id, idSize}
				,
				{CKA_SUBJECT, subject, subjectSize}
				,
				{CKA_LABEL, label, labelSize}
			};

			DEBUG("ID size: %ld\n", idSize);

			rv = C_SetAttributeValue(globalState.shSession, hObject,
				pValueTemplate,
				sizeof(pValueTemplate) / sizeof(CK_ATTRIBUTE));
		}

		if (rv != CKR_OK)
		{
			DEBUG("Could not set CKA_ID for private key, error: %s (0x%08lX)\n",
				lookup_enum(RV_T, rv), rv);
			goto import_err;
		}
		// start search for public keys
		if ((rv = C_FindObjectsInit(globalState.shSession, NULL, 0)) != CKR_OK)
		{
			DEBUG("Could not initialize object search, error: %s (0x%08lX)\n",
				lookup_enum(RV_T, rv), rv);
			goto import_err;
		}

		newPubKeyNum = 0;
		while (TRUE)
		{
			// do the search
			if ((rv = C_FindObjects(globalState.shSession, ohObjects, 1,
						&ulNumFound)) != CKR_OK)
			{
				DEBUG("Could not perform the FindObjects, error %s (0x%08lX)\n",
					lookup_enum(RV_T, rv), rv);
				goto import_err;
			}
			if (ulNumFound == 0)	// No more objects to read
				break;

			// Get the object class
			pValueTemplate[0].type = CKA_CLASS;
			pValueTemplate[0].pValue = &ocClass;
			pValueTemplate[0].ulValueLen = sizeof(ocClass);

			if ((rv = C_GetAttributeValue(globalState.shSession,
						ohObjects[0], pValueTemplate, 1)) != CKR_OK)
			{
				DEBUG("Error: Could not get the objects class: %s (0x%08lX)\n",
					lookup_enum(RV_T, rv), rv);
				continue;
			}

			if (ocClass == CKO_PUBLIC_KEY)
			{
				newPubKeys[newPubKeyNum++] = ohObjects[0];
			}
		}

		C_FindObjectsFinal(globalState.shSession);

		DEBUG("Number of old keys: %d\tNew Keys: %d\n", oldPubKeyNum,
			newPubKeyNum);

		for (i = 0; i < newPubKeyNum; i++)
		{
			int j;
			newPubKey = newPubKeys[i];
			for (j = 0; j < oldPubKeyNum; j++)
			{
				if (newPubKey == oldPubKeys[j])
					break;
			}
			DEBUG("j = %d\toldPubKeyNum: %d\n", j, oldPubKeyNum);
			if (j == oldPubKeyNum)
				break;
		}
		DEBUG("i = %d\tnewPubKeyNum: %d\n", i, newPubKeyNum);
		if (i != newPubKeyNum)
		{
			CK_ATTRIBUTE pValueTemplate = { CKA_ID, id, idSize };

			DEBUG("ID size: %ld\n", idSize);

			rv = C_SetAttributeValue(globalState.shSession,
				newPubKey, &pValueTemplate, 1);

			if (rv != CKR_OK)
				DEBUG("Could not set CKA_ID for public key, error: %s (0x%08lX)\n", lookup_enum(RV_T, rv), rv);
		}

		readKeyInfo(hObject, CKO_PRIVATE_KEY);

		if (i != newPubKeyNum)
		{
			readKeyInfo(newPubKey, CKO_PUBLIC_KEY);
		}

	}

  import_err:
	PKCS12_free(p12);
	EVP_PKEY_free(pkey);
	RSA_free(key);
	free(subject);
	free(id);
	free(modulus);
	free(publicExponent);
	free(privateExponent);
	free(prime1);
	free(prime2);
	free(exponent1);
	free(exponent2);
	free(coefficient);
	free(label);

	return (rv == CKR_OK);
}

/* Create new data object from a file */
BOOL createDataObject(gchar * fileName)
{
	FILE *fp;
	struct stat buf;
	CK_OBJECT_HANDLE hData;
	CK_OBJECT_CLASS dataClass = CKO_DATA;
	CK_CHAR application[] = "GTKCard";
	CK_BYTE_PTR dataValue;
	CK_BYTE true = TRUE;

	CK_RV rv;

	DEBUG("Importing: %s\n", fileName);

	// Open file
	fp = fopen(fileName, "rb");
	if (fp == NULL)
	{
		DEBUG("fopen on %s failed: %s\n", fileName, strerror(errno));
		return FALSE;
	}

	fstat(fileno(fp), &buf);

	DEBUG("File opened successfully, size %ld\n", buf.st_size);

	dataValue = malloc(buf.st_size);

	if (buf.st_size != fread(dataValue, 1, buf.st_size, fp))
	{
		DEBUG("fread on %s failed: %s\n", fileName, strerror(errno));
		fclose(fp);
		return FALSE;
	}

	fclose(fp);

	/* Create a data object */
	CK_ATTRIBUTE dataTemplate[] = {
		{CKA_CLASS, &dataClass, sizeof(dataClass)}
		,
		{CKA_TOKEN, &true, sizeof(true)}
		,
		{CKA_APPLICATION, application, sizeof(application) }
		,
		{CKA_VALUE, dataValue, buf.st_size}
	};

	rv = C_CreateObject(globalState.shSession, dataTemplate,
		sizeof(dataTemplate)/sizeof(dataTemplate[0]), &hData);

	free(dataValue);

	if (rv != CKR_OK)
	{
		char string[200];
		snprintf(string, 200, "Could not create data object: %s (0x%08lX)",
			lookup_enum(RV_T, rv), rv);
		DELETE();
		DISPLAY(string);
		return FALSE;
	}

	readObjectInfo(hData);

	DELETE();
	DISPLAY("Object created successfully\n");

	return TRUE;
}

BOOL P11_ReadObject(CK_OBJECT_HANDLE ohObject, gchar ** text)
{
	// P11 variables
	CK_ATTRIBUTE pValueTemplate[1];
	CK_RV rv;
	CK_ULONG size;

	pValueTemplate[0].type = CKA_VALUE;
	pValueTemplate[0].pValue = NULL;
	pValueTemplate[0].ulValueLen = 0;

	ASSERT((rv = C_GetAttributeValue(globalState.shSession, ohObject,
				pValueTemplate, 1)) == CKR_OK,
		"Could not get the objects size", rv);

	size = pValueTemplate[0].ulValueLen;
	*text = malloc(size);

	pValueTemplate[0].type = CKA_VALUE;
	pValueTemplate[0].pValue = *text;
	pValueTemplate[0].ulValueLen = size;

	ASSERT((rv = C_GetAttributeValue(globalState.shSession, ohObject,
				pValueTemplate, 1)) == CKR_OK,
		"Could not get the objects attribute values", rv);

	return TRUE;
}

BOOL P11_displayCert(pCertInfo certInfo, GtkWidget * textBox)
{
	// P11 variables
	CK_ATTRIBUTE pValueTemplate[1];
	CK_RV rv;
	CK_BYTE tmpCert[certInfo->size];
	char string[1000];

	// OpenSSL variables
	X509 *cert = 0;
	BIO *mem = BIO_new(BIO_s_mem());
	STACK *nlist;
	ASN1_TIME *time;
	ASN1_INTEGER *integer;
	X509_NAME *name;
	int len;
	int j;
	unsigned int n;
	char buffer[1000];
	unsigned char md[EVP_MAX_MD_SIZE];
	const EVP_MD *md_alg, *digest = EVP_md5();

	if (textBox == NULL)
	{
		DEBUG("NULL\n");
		return FALSE;
	}

	pValueTemplate[0].type = CKA_VALUE;
	pValueTemplate[0].pValue = tmpCert;
	pValueTemplate[0].ulValueLen = certInfo->size;

	ASSERT((rv =
			C_GetAttributeValue(globalState.shSession, certInfo->ohHandle,
				pValueTemplate, 1)) == CKR_OK,
		"Could not get the certificates attribute values", rv);

	BIO_write(mem, tmpCert, certInfo->size);

	d2i_X509_bio(mem, &cert);

	if (cert == NULL)
		return FALSE;

	if ((md_alg = EVP_get_digestbyname("sha1")))
	{
		/* ok */
		DEBUG("Sha1\n");
		digest = md_alg;
	}

	integer = X509_get_serialNumber(cert);
	for (j = 0; j < integer->length; j++)
		sprintf(&buffer[j * 3], "%02X%c", integer->data[j],
			(j + 1 == (int) integer->length) ? '\0' : ':');
	snprintf(string, 1000, "SerialNumber:\t\t %s\n", buffer);
	DISPLAY_FIXED(string);

	name = X509_get_subject_name(cert);
	nlist = (STACK *) name->entries;
	X509_NAME_print(mem, name, 0);
	len = BIO_read(mem, buffer, 1000);
	buffer[len] = '\0';
	snprintf(string, 1000, "UserName:\t\t %s\n", buffer);
	DISPLAY_FIXED(string);

	name = X509_get_issuer_name(cert);
	nlist = (STACK *) name->entries;
	X509_NAME_print(mem, name, 0);
	len = BIO_read(mem, buffer, 1000);
	buffer[len] = '\0';
	snprintf(string, 1000, "Issuer:\t\t\t %s\n", buffer);
	DISPLAY_FIXED(string);

	time = X509_get_notBefore(cert);
	ASN1_TIME_print(mem, time);
	len = BIO_read(mem, buffer, 100);
	buffer[len] = '\0';
	snprintf(string, 1000, "Valid From:\t\t %s\n", buffer);
	DISPLAY_FIXED(string);

	time = X509_get_notAfter(cert);
	ASN1_TIME_print(mem, time);
	len = BIO_read(mem, buffer, 100);
	buffer[len] = '\0';
	snprintf(string, 1000, "Valid Until:\t\t %s\n", buffer);
	DISPLAY_FIXED(string);

	X509_digest(cert, digest, md, &n);

	sprintf(buffer, "FingerPrint(%s):\t ", OBJ_nid2sn(EVP_MD_type(digest)));
	DISPLAY_FIXED(buffer);

	for (j = 0; j < (int) (n); j++)
	{
		sprintf(&buffer[j * 3], "%02X%c", md[j],
			(j + 1 == (int) n) ? '\n' : ':');
	}

	DISPLAY_FIXED(buffer);

	BIO_vfree(mem);
	X509_free(cert);

	return TRUE;
}

/***********************************************************************/
/*                         Utility Functions                           */
/***********************************************************************/

// Method to read a certificate from a file
BOOL getFileCert(const char *file, X509 ** cert)
{
	FILE *fp;
	X509 *mycert = 0;

	// Open file
	fp = fopen(file, "rb");
	if (fp == NULL)
		return FALSE;

	// Get the cert
	d2i_X509_fp(fp, &mycert);
	if (mycert == NULL)
		return FALSE;

	fclose(fp);

	*cert = mycert;

	return TRUE;
}

// Method to read a key from a PKCS#12 file
BOOL getFileP12(const char *file, PKCS12 ** p12)
{
	FILE *fp;
	PKCS12 *myP12 = 0;

	// Open file
	fp = fopen(file, "rb");
	if (fp == NULL)
		return FALSE;

	// Get the cert
	d2i_PKCS12_fp(fp, &myP12);
	if (myP12 == NULL)
		return FALSE;

	fclose(fp);

	*p12 = myP12;

	return TRUE;
}

// Local method to import a certificate to the card
BOOL localImportCert(X509 * cert, gchar * certID, CK_OBJECT_HANDLE * ohHandle)
{
	CK_RV rv;
	int rvi;
	CK_BYTE_PTR subject = NULL_PTR;
	CK_BYTE_PTR certificateValue = NULL_PTR;
	CK_BYTE_PTR tmp;
	CK_BYTE true = TRUE;
	CK_OBJECT_HANDLE hCertificate = 0;
	CK_OBJECT_CLASS certificateClass = CKO_CERTIFICATE;
	int certSize, subjectSize, issuerSize;
	CK_ULONG certType = CKC_X_509;
	STACK *nlist;
	X509_NAME *name;
	int len;
	CK_BYTE_PTR issuer;
	CK_BYTE subjectBuf[256];

	BIO *mem = BIO_new(BIO_s_mem());

	if (!(subjectSize = i2d_X509_NAME(cert->cert_info->subject, NULL)))
	{
		DEBUG("Could not DER encode subject %d\n", subjectSize);
		X509_free(cert);
		return FALSE;
	}

	subject = malloc(subjectSize);
	tmp = subject;

	if (!(rvi = i2d_X509_NAME(cert->cert_info->subject, &tmp)))
	{
		DEBUG("Could not DER encode subject %d\n", rvi);
		X509_free(cert);
		free(subject);
		return FALSE;
	}

	if (!(issuerSize = i2d_X509_NAME(cert->cert_info->issuer, NULL)))
	{
		DEBUG("Could not DER encode issuer %d\n", issuerSize);
		X509_free(cert);
		return FALSE;
	}

	issuer = malloc(issuerSize);
	tmp = issuer;

	if (!(rvi = i2d_X509_NAME(cert->cert_info->issuer, &tmp)))
	{
		DEBUG("Could not DER encode issuer %d\n", rvi);
		X509_free(cert);
		free(subject);
		free(issuer);
		return FALSE;
	}

	name = X509_get_subject_name(cert);
	nlist = (STACK *) name->entries;
	X509_NAME_print(mem, name, 0);
	len = BIO_read(mem, subjectBuf, 256);
	subjectBuf[len] = '\0';
	BIO_vfree(mem);

	if (!(certSize = i2d_X509(cert, NULL)))
	{
		DEBUG("Could not DER encode the certificate %d\n", certSize);
		X509_free(cert);
		free(subject);
		free(issuer);
		return FALSE;
	}

	certificateValue = malloc(certSize);
	tmp = certificateValue;

	if (!(rv = i2d_X509(cert, &tmp)))
	{
		DEBUG("Could not DER encode the certificate %s (0x%08lX)\n",
			lookup_enum(RV_T, rv), rv);
		X509_free(cert);
		free(subject);
		free(certificateValue);
		free(issuer);
		return FALSE;
	}

	CK_ATTRIBUTE certificateTemplate[] = {
		{CKA_CLASS, &certificateClass, sizeof(certificateClass)}
		,
		{CKA_TOKEN, &true, sizeof(true)}
		,
		{CKA_SUBJECT, subject, subjectSize}
		,
		{CKA_LABEL, subjectBuf, len}
		,
		{CKA_VALUE, certificateValue, certSize}
		,
		{CKA_CERTIFICATE_TYPE, &certType, sizeof(certType)}
		,
		{CKA_ISSUER, issuer, issuerSize}
	};

	if ((rv = C_CreateObject(globalState.shSession, certificateTemplate,
				sizeof(certificateTemplate) /
				sizeof(CK_ATTRIBUTE), &hCertificate)) != CKR_OK)
	{
		DEBUG("Could not create object: %s (0x%08lX)\n",
			lookup_enum(RV_T, rv), rv);
		X509_free(cert);
		free(subject);
		free(certificateValue);
		free(issuer);
		return FALSE;
	}

	if (certID != NULL)
	{
		CK_ATTRIBUTE pValueTemplate = { CKA_ID, certID, strlen(certID) };

		rv = C_SetAttributeValue(globalState.shSession, hCertificate,
			&pValueTemplate, 1);

		if (rv != CKR_OK)
			DEBUG("Could not set CKA_ID for the certificate, error: %s (0x%08lX)\n", lookup_enum(RV_T, rv), rv);
	}
	else
	{							// Set the CKA_ID to be the hash of the modulus
		EVP_PKEY *pub_key = (EVP_PKEY *) X509_get_pubkey(cert);
		BIGNUM *n = pub_key->pkey.rsa->n;
		CK_BYTE hash[20];
		unsigned int hashLen;
		CK_ULONG modulusSize = BN_num_bytes(n);
		CK_BYTE_PTR modulus = malloc(modulusSize);
		BN_bn2bin(n, modulus);

		const EVP_MD *m = EVP_sha1();
		EVP_MD_CTX ctx;

		EVP_DigestInit(&ctx, m);
		EVP_DigestUpdate(&ctx, modulus, modulusSize);
		EVP_DigestFinal(&ctx, hash, &hashLen);

		CK_ATTRIBUTE pValueTemplate = { CKA_ID, hash, hashLen };

		rv = C_SetAttributeValue(globalState.shSession, hCertificate,
			&pValueTemplate, 1);

		if (rv != CKR_OK)
			DEBUG ("Could not set CKA_ID for the certificate, error: %s (0x%08lX)\n", lookup_enum(RV_T, rv), rv);

		free(modulus);
	}

	if (ohHandle != NULL)
		*ohHandle = hCertificate;

	X509_free(cert);
	free(subject);
	free(certificateValue);
	free(issuer);

	return TRUE;
}

CK_BYTE *keyType2Str(CK_KEY_TYPE keyType)
{
	const int MAX_LEN = 11;
	char *keyString = malloc(MAX_LEN * sizeof(CK_BYTE));

	switch (keyType)
	{
	case CKK_RSA:
		snprintf(keyString, MAX_LEN, "RSA");
		break;
	case CKK_DSA:
		snprintf(keyString, MAX_LEN, "DSA");
		break;
	case CKK_DES:
		snprintf(keyString, MAX_LEN, "DES");
		break;
	case CKK_DES2:
		snprintf(keyString, MAX_LEN, "3DES 2 KEY");
		break;
	case CKK_DES3:
		snprintf(keyString, MAX_LEN, "3DES");
		break;
	default:
		snprintf(keyString, MAX_LEN, "N/A");
		break;
	}

	return (CK_BYTE *)keyString;
}
